using System;
using System.Collections.Generic;
using ceometric.VectorGeometry;
using ceometric.ComputationalGeometry;

//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// You must reference both V# and G# first to run this program
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

namespace CuttingTool
{
    /// <summary>
    /// A program to compute the path of a circular cutting tool with radius r running
    /// along an arbitrary convex or concave edge of a workpiece so that the cutting tool
    /// touches the workpiece at least in one point.
    /// The workpiece must only consist of straight edges. The computed path of the 
    /// cutting tool will also solely consist of edges. Any resulting arcs in the path
    /// will be broken down into a specified numer of edges.
    /// </summary>
    class Program
    {
        /// <summary>Defines on which side of the workpiece the tool runs. Set eighter -1 or 1.</summary>
        public static int dir = 1;

        static void Main(string[] args)
        {
            //-----------------------------------------------------------------------------------------
            List<Edge> path = new List<Edge>(); //Holds the computed path of the cutting tool           
            double r = 0.6; //The radius of the cutting tool

            //-----------------------------------------------------------------------------------------
            //Create a list of edges representing the workpiece boundary
            //Create a list of oriented edges. Note that for the sake of ease
            //1. the end point of an edge must be the start point of the next
            //2. edges must not be collinear 
            //3. edges must not have zero-length
            //or this method will fail.
            Point p1 = new Point(0, -5, 0);
            Point p2 = new Point(0, 5, 0);
            Point p3 = new Point(2, 6, 0);
            Point p4 = new Point(6, 6, 0);
            Point p5 = new Point(7, 7, 0);
            Point p6 = new Point(7, 9, 0);
            Point p7 = new Point(7.5, 9.5, 0);
            Point p8 = new Point(8, 9, 0);
            Point p9 = new Point(10, 5, 0);
            Point p10 = new Point(11, 5, 0);
            Point p11 = new Point(9, 9, 0);
            Point p12 = new Point(9.5, 9, 0);
            Point p13 = new Point(9.5, 10, 0);
            Point p14 = new Point(9.5, 12, 0);

            List<Edge> boundary = new List<Edge>();
            boundary.Add(new Edge(p1, p2));
            boundary.Add(new Edge(p2, p3));
            boundary.Add(new Edge(p3, p4));
            boundary.Add(new Edge(p4, p5));
            boundary.Add(new Edge(p5, p6));
            boundary.Add(new Edge(p6, p7));
            boundary.Add(new Edge(p7, p8));
            boundary.Add(new Edge(p8, p9));
            boundary.Add(new Edge(p9, p10));
            boundary.Add(new Edge(p10, p11));
            boundary.Add(new Edge(p11, p12));
            boundary.Add(new Edge(p12, p13));
            boundary.Add(new Edge(p13, p14));

            //A time gauge
            KernelTimeGauge tg = new KernelTimeGauge();
            tg.Start();

            //-----------------------------------------------------------------------------------------
            //Compute tool path
            //First, compute the normal vectors for the edges. This will help determining on which side
            //of an edge we are. A normal vector always points into the direction of the cutting tool.
            //Let's say that the cutting tool runs on the right side of the edges as in the example. This is,
            //the triangle startpoint->endpoint->center_cutting_tool is oriented clockwise when viewed from above.  
            //The normal vector can be computed using the cross product (right hand rule!). 
            List<Vector3d> normalVectors = new List<Vector3d>();
            Vector3d nv = new Vector3d(0, 0, dir);
            for ( int i = 0; i < boundary.Count; i++ )
            {
                Vector3d n = Vector3d.Cross(boundary[i].ToVector3d(), nv);
                //Setting the length of the normal vector equal to the radius of the cutting tool will help us later
                n.Norm = r;
                normalVectors.Add(n);
            }

            //Second, offset the boundary in the direction of its normal vector.
            List<Edge> offsettedBoundary = new List<Edge>();
            for ( int i = 0; i < boundary.Count - 1; i++ )
            {
                offsettedBoundary.Add(boundary[i].Move(normalVectors[i]));
                //If left turn, add the segment in addition
                if (dir < 0)
                {
                    if (Predicate.Orient2d(boundary[i].StartPoint, boundary[i].EndPoint, boundary[i + 1].EndPoint) < 0)
                    {
                        offsettedBoundary.AddRange(Segment(normalVectors[i], normalVectors[i + 1], boundary[i].EndPoint));
                    }
                }
                else
                {
                    if (Predicate.Orient2d(boundary[i].StartPoint, boundary[i].EndPoint, boundary[i + 1].EndPoint) > 0)
                    {
                        offsettedBoundary.AddRange(Segment(normalVectors[i], normalVectors[i + 1], boundary[i].EndPoint));
                    }
                }
            }
            //add the last edge
            offsettedBoundary.Add(boundary[boundary.Count - 1].Move(normalVectors[boundary.Count - 1]));

            //This is the tricky part:
            //For an offsetted edge i we count from the maximum number of offsetted edges down to i intersecting
            //the edge j with the actual edge i:
            //If the offsetted edge i intersects with j, then take the intersection point as the end point of 
            //path i and start point of the next path. The point is that this intrinsically handles the odd case
            //that a boundary line is too short, or, in other words, that the radius of the cutting tool is too
            //large to reach into the edge of the boundary (situation at the upper right corner in your example). 

            List<Circle> circles = new List<Circle>(); //just for visualization
            Point startPoint = offsettedBoundary[0].StartPoint;
            circles.Add(new Circle(startPoint, r, new Vector3d(0, 0, 1)));
            for ( int i = 0; i < offsettedBoundary.Count; i++ )
            {
                //Indices of intersection edges
                List<int> intersectionIndices = new List<int>();

                for (int j = offsettedBoundary.Count - 1; j >= i; j--)
                {
                    //memorize intersection points
                    if (Predicate.IntersectsExact(offsettedBoundary[i], offsettedBoundary[j]) == 1) intersectionIndices.Add(j);
                }

                if ( intersectionIndices.Count == 0 )
                {
                    path.Add(new Edge(startPoint, offsettedBoundary[i].EndPoint));
                    if (i < offsettedBoundary.Count-1)
                    {
                        startPoint = offsettedBoundary[i + 1].StartPoint;
                    }
                    else
                    {
                        startPoint = offsettedBoundary[i].EndPoint;
                    }
                    circles.Add(new Circle(startPoint, r, new Vector3d(0, 0, 1)));
                }
                else
                {
                    //identify the intersection point that lies closest to the current start point
                    Point closestPoint = new Point();
                    int closestIndex = int.MaxValue;
                    double distance = double.MaxValue;
                    for ( int j = 0; j < intersectionIndices.Count; j++ )
                    {
                        int index = intersectionIndices[j];
                        Point ip = offsettedBoundary[i].Intersect3d(offsettedBoundary[index]);
                        if ( ip.DistanceTo(startPoint) < distance )
                        {
                            distance = ip.DistanceTo(startPoint);
                            closestPoint = ip;
                            closestIndex = index;
                        }
                    }
                    path.Add(new Edge(startPoint, closestPoint));
                    startPoint = closestPoint;
                    i = closestIndex - 1;
                    circles.Add(new Circle(closestPoint, r, new Vector3d(0, 0, 1)));
                }
            }

            tg.Stop();
            Console.WriteLine("Time to compute tool path: " + tg.ToString());

            //---------------------------------------------------------------------
            // Visualize in dxf format
            // Get the free DWG TrueView viewer at www.autodesk.com to view dxf
            Scene sc = new Scene();

            Scene.Layer pathLayer = new Scene.Layer("tool_path", 2);
            Scene.Layer boundaryLayer = new Scene.Layer("workpiece_boundary", 7);
            Scene.Layer toolLayer = new Scene.Layer("tool", 4);

            sc.Add(circles, toolLayer);
            sc.Add(path, pathLayer);
            sc.Add(boundary, boundaryLayer);

            string outfile = @"..\..\cutting_path.dxf";
            sc.WriteDxf(outfile);
            Console.WriteLine("dxf written to " + outfile);

            Console.WriteLine("Hit any key to continue.");
            Console.ReadLine();
        }

        public static List<Edge> Segment(Vector3d n1, Vector3d n2, Point center)
        {
            //A list holding the segment edges
            List<Edge> segment = new List<Edge>();

            //Let a segment consist of n edges
            int n = 64;

            //Caution! works only for angles less than 90 deg. Otherwise, make a case differentiation!
            double ange = dir * Vector3d.Angle(n1, n2);
            double deltaAngle = ange/n;
            
            //A list holding the points on the segment
            List<Point> pointsOnSegment = new List<Point>();
            for ( int i = 0; i <= n; i++ )
            {
                //The points are generated by rotating the normal vector around the center point
                Matrix3d rMat = Matrix3d.RotationZAxis((double)i * deltaAngle);
                pointsOnSegment.Add(center + n1.Rotate(rMat).ToPoint());
            }

            //Connect the points to get the segment edges
            for ( int i = 1; i < pointsOnSegment.Count; i++ )
            {
                segment.Add(new Edge(pointsOnSegment[i - 1], pointsOnSegment[i]));
            }

            return segment;
        }
    }
}