using System;
using System.Collections.Generic;
using ceometric.ComputationalGeometry;
using ceometric.VectorGeometry;

namespace ProjectLinesOnSurface
{
    class Program
    {
        static void Main(string[] args)
        {
            //---------------------------------------------------------------------
            //Read some points from a sample point set in xyz format.
            //SPECIFY THE CORRECT FILE FORMAT HERE:
            //terrain_en.xyz has comma (",") decimal seperator
            //terrain_us.xyz has point (".") decimal seperator
            PointSet vertices = new PointSet();
            //vertices.ReadFromFile(@"..\..\terrain_us.xyz", ';');
            vertices.ReadFromFile(@"..\..\terrain_en.xyz", ';');

            //---------------------------------------------------------------------
            //Triangulate the pure xyz point data.
            Console.Write("Triangulating...");
            DelaunayTriangulation2d dt = new DelaunayTriangulation2d();
            dt.TriangulationPoints = vertices;
            dt.Triangulate();
            List<Triangle> triangles = dt.GetTriangles(); //Get the triangles of the surface
            Console.WriteLine("done.");

            //---------------------------------------------------------------------
            //Build some edges to project
            Point p1 = new Point(480780, 4401270, 7000);
            Point p2 = new Point(486030, 4401270, 7000);
            Point p3 = new Point(486030, 4404330, 7000);
            Point p4 = new Point(480780, 4404330, 7000);
            List<Edge> edges = new List<Edge>();
            edges.Add(new Edge(p1, p2));
            edges.Add(new Edge(p2, p3));
            edges.Add(new Edge(p3, p4));
            edges.Add(new Edge(p4, p1));

            //---------------------------------------------------------------------
            //Project edges onto triangulated surface
            Console.WriteLine();
            KernelTimeGauge tg = new KernelTimeGauge();
            tg.Start();
            EdgeSurfaceProjection esp = new EdgeSurfaceProjection();
            esp.EliminateZeroLengthEdges = false;
            List<Edge> projectedEdges = esp.Project(edges, triangles, new Vector3d(0, 0, 1));
            tg.Stop();
            Console.WriteLine(edges.Count + " edges projected onto " + triangles.Count + " triangles in " + tg.DurationInMilliSeconds + "ms");
            Console.WriteLine("Number of projected edges: " + projectedEdges.Count);

            //---------------------------------------------------------------------
            //Create a polyline from the projected edges. Note that the polyline construction is a static
            //method since more than one polyline may result for a set of unordered edges in a general case.
            //Decrease the number of relevant digits if the edges are disconnected and more than one polyline
            //is created when it shouldn't.
            int numberOfZeroLengthEdges = 0;
            int relevantDigits = 15;
            bool eliminateZeroLengthSegments = true;
            List<Polyline> pls = Polyline.ConstructFromUnorderedSegments(
                projectedEdges, relevantDigits, ref numberOfZeroLengthEdges, eliminateZeroLengthSegments);
            Polyline pl = pls[0]; //If not all edges are connected or if there are branches, there will be more than one polyline.

            //---------------------------------------------------------------------
            //Polyline simplification:
            //Compute a polyline where the maximum distance of each vertex of the
            //original polyline to the new polyline is smaller than specified.
            Polyline spl = pl.Simplify(1e-4);
            Console.WriteLine("Number of simplified edges: " + spl.Count);

            //---------------------------------------------------------------------
            // Visualize in dxf format
            // Get the free DWG TrueView viewer at www.autodesk.com to view dxf
            Console.WriteLine();
            Scene sc = new Scene();
            Scene.Layer eLayer = new Scene.Layer("Edges", 1);
            Scene.Layer peLayer = new Scene.Layer("Projected_edges", 5);
            Scene.Layer speLayer = new Scene.Layer("Simplified_edges", 4);
            Scene.Layer tLayer = new Scene.Layer("Delaunay_Triangles", 3);

            // Add objects to the scene
            sc.Add(edges, eLayer);
            sc.Add(triangles, tLayer);
            sc.Add(projectedEdges, peLayer);
            sc.Add(spl, speLayer);

            // Output the scene as dxf
            string outfile = @"..\..\edge_projection.dxf";
            sc.WriteDxf(outfile);
            Console.WriteLine("dxf written to " + outfile);

            Console.WriteLine("Hit any key to continue.");
            Console.ReadLine();
        }
    }
}